#include <sta/drivers/MS56xx.hpp>

#include <sta/debug/assert.hpp>
#include <sta/lang.hpp>

#include <cmath>


namespace sta 
{
    MS56xx::MS56xx(SPIDevice * device, DelayUsFunc delay, OsrLevel level /* = OsrLevel::_1024 */)
        : device_{device},
          delay_{delay},
          osr_{level},
          intf_{Intf::SPI}
    {
        STA_ASSERT(device != nullptr);
    }

    MS56xx::MS56xx(I2CDevice * device, DelayUsFunc delay, OsrLevel osr /* = OsrLevel::_1024 */)
        : device_{device},
          delay_{delay},
          osr_{osr},
          intf_{Intf::I2C}
    {
        STA_ASSERT(device != nullptr);
    }

    bool MS56xx::init()
    {
        // Reset device on start-up
        this->reset();

        // Read the PROM e.g. constants
        this->readPROM();
    }

    void MS56xx::setOsr(OsrLevel osr)
    {
        osr_ = osr;
    }

    void MS56xx::reset()
    {
        busCommand(MS56xx::Operations::RESET);
        delay_(MS56xx::RESET_DELAY);
    }

    float MS56xx::getPressure(Unit unit /* = Unit::hPa */)
    {
        // Request the ADC to read temperature values.
        busCommand(MS56xx::Operations::D2_CONVERSION + 2*this->osr_);

        // 8.22 ms conversion according to the datasheet.
        delay_(osrDelay());

        uint8_t buffer[3] = { 0x00, 0x00, 0x00 };
        busRead(MS56xx::Operations::ADC_RESULT, buffer, 3);

        // Difference between actual and reference temperature.
        uint32_t rawTemperature = buffer[0] << 16 | buffer[1] << 8 | buffer[2];
        uint32_t dT = rawTemperature - refTemp_ * (0x01<<8);

        // Request the ADC to read pressure values.
        busCommand(MS56xx::Operations::D1_CONVERSION + 2*this->osr_);

        // 8.22 ms conversion according to the datasheet.
        delay_(osrDelay());

        busRead(MS56xx::Operations::ADC_RESULT, buffer, 3);

        uint32_t rawPressure = buffer[0] << 16 | buffer[1] << 8 | buffer[2];

        // offset and sensitivity at actual temperature.
        uint64_t offset = offset_ * (0x01<<16) + (tempOffsetCoeff_ * dT) / (0x01<<7);
        uint64_t sensibility = sensibility_ * (0x01<<15) + (tempSensCoeff_ * dT) / (0x01<<8); 

        // The pressure in Pa.
        float pressure = (rawPressure * (sensibility / 0x01<<21) - offset) / (0x01<<15);
        pressure = convertPressure(pressure, unit);

        return pressure;
    }

    int32_t MS56xx::getTemperature() {
        busCommand(MS56xx::Operations::D2_CONVERSION + 2*this->osr_);

        // 8.22 ms conversion according to the datasheet.
        delay_(osrDelay());

        uint8_t buffer[3] = { 0x00, 0x00, 0x00 };
        busRead(MS56xx::Operations::ADC_RESULT, buffer, 3);
        uint32_t rawTemperature = buffer[0] | buffer[1] << 8 | buffer[2] << 16;

        // Convert the raw values to actual temperature values.
        uint32_t dT = rawTemperature - refTemp_ * (0x01<<8);
        uint32_t temperature = 20 * dT * tempCoeff_ / (0x01<<23);

        return temperature;
    }

    void MS56xx::setPressureReference(float pressRef, float altRef, Unit unit /* = Unit::hPa */)
    {
        // Taken from: https://cdn-shop.adafruit.com/datasheets/BST-BMP180-DS000-09.pdf, page 17.
        sealevel_ = pressRef / (std::pow((1 - altRef / 44330), 5.255));
    }

    float MS56xx::getAltitudeEstimate()
    {
        float pressure = getPressure(Unit::hPa);

        // Taken from: https://cdn-shop.adafruit.com/datasheets/BST-BMP180-DS000-09.pdf, page 16
        float altitude = 44330 * (1 - std::pow(pressure / sealevel_, 1 / 5.255));
    }

    void MS56xx::readPROM() {
        uint8_t buffer[2];

        busRead(Operations::READ_PROM, buffer, 2);
        sensibility_ = uint_8BufferTouint16_t(buffer);

        delay_(1000);

        busRead(Operations::READ_PROM+2, buffer, 2);
        offset_ = uint_8BufferTouint16_t(buffer);

        delay_(1000);

        busRead(Operations::READ_PROM+4, buffer, 2);
        tempSensCoeff_ = uint_8BufferTouint16_t(buffer);

        delay_(1000);

        busRead(Operations::READ_PROM+6, buffer, 2);
        tempOffsetCoeff_ = uint_8BufferTouint16_t(buffer);

        delay_(1000);

        busRead(Operations::READ_PROM+8, buffer, 2);
        refTemp_ = uint_8BufferTouint16_t(buffer);

        delay_(1000);

        busRead(Operations::READ_PROM+10, buffer, 2);
        tempCoeff_ = uint_8BufferTouint16_t(buffer);
    }

    inline float MS56xx::convertPressure(float pressure, Unit unit)
    {
        switch (unit)
        {
        case Unit::mbar:
            return pressure / 100.0f;
        case Unit::hPa:
            return pressure / 100.0f;
        case Unit::Pa:
            return pressure;
        default:
            STA_UNREACHABLE();
            break;
        }
    }

    inline uint16_t MS56xx::uint_8BufferTouint16_t(uint8_t* buffer) {
        return (buffer[0] << 8) | buffer[1];
    }

    bool MS56xx::busCommand(uint8_t command)
    {
        this->device_->beginTransmission();
        this->device_->transfer(command);
        this->device_->endTransmission();

        return true;
    }

    uint32_t MS56xx::osrDelay() {
        // Delay times taken from:
        // https://www.amsys-sensor.com/downloads/notes/MS5XXX-C-code-example-for-MS56xx-MS57xx-MS58xx-AMSYS-an520e.pdf
        switch (osr_) {
        case _256:
            return 900;
        case _512:
            return 3000;
        case _1024:
            return 4000;
        case _2048:
            return 6000;
        case _4096:
            return 10000;
        }
    }

    bool MS56xx::busRead(uint8_t reg, uint8_t * buffer, size_t length)
    {
        this->device_->beginTransmission();
        this->device_->transfer(reg);
        this->device_->receive(buffer, length);
        this->device_->endTransmission();

        return true;
    }

    bool MS56xx::busWrite(uint8_t reg, uint8_t * buffer, size_t length)
    {
        return true;
    }
}